﻿using PTAPSWebServiceClientSample.APSWebServiceRef;
using System;
using System.Collections.Generic;
using System.IO;
using System.Linq;
using System.ServiceModel;
using System.Text;
using System.Threading.Tasks;

namespace PTAPSWebServiceClientSample
{
    //Current Contract values
    /*
    public enum EApsWebServicesResponseCodes
    {
        Success = 0,
        SuccessWihoutValidation = 1,
        InvalidDateFormat = 2,
        InvalidDatePast = 3,
        InvalidDateFuture = 4,
        InvalidUserCredentials = 5,
        InvlaidUserPermissions = 6,
        NoServiceListening = 7,
        NoInstanceWithThatName = 8,
        ValidationTimeout = 9,
        NoServerManager = 10,
        InvalidScenarioName = 11,
        FailedToBroadcast = 12,
        ProcessingTimeout = 13,
        Failure = 14,
    }

    [ServiceContract]
    public interface IExtraServicesWebService
    {
        [OperationContract]
        EApsWebServicesResponseCodes AdvanceClock(string a_user, string a_password, string a_dateTime);

        [OperationContract]
        EApsWebServicesResponseCodes AdvanceClockTicks(string a_user, string a_password, long a_dateTimeTicks);

        [OperationContract]
        EApsWebServicesResponseCodes PublishLive(string a_user, string a_password, int a_publishType);

        [OperationContract]
        EApsWebServicesResponseCodes PublishByName(string a_user, string a_password, int a_publishType, string a_scenarioName);

        [OperationContract]
        Task<ApsWebServiceResponseBase> Import(ImportRequest a_request);

        [OperationContract]
        ApsWebServiceResponseBase Optimize(OptimizeRequest a_request);

        [OperationContract]
        Task<ApsWebServiceResponseBase> Publish(PublishRequest a_request);
        
        [OperationContract]
        CtpResponse CTP(string a_userName, string a_password, CtpRequest a_ctpRequest);
    }
    */

    class Program
    {
        /// <summary>
        /// For individual testing of service calls - uncomment related section
        /// </summary>
        /// <param name="args"></param>
        static void Main(string[] args)
        {
            //Setup the binding. This is done in code so the config file is not needed
            BasicHttpBinding binding = new BasicHttpBinding(BasicHttpSecurityMode.None);
            binding.BypassProxyOnLocal = true;
            binding.CloseTimeout = new TimeSpan(0, 2, 0);
            binding.OpenTimeout = new TimeSpan(0, 0, 30);
            binding.SendTimeout = new TimeSpan(0, 10, 0);
            binding.AllowCookies = false;
            binding.HostNameComparisonMode = HostNameComparisonMode.StrongWildcard;
            binding.MaxBufferPoolSize = 524288;
            binding.MaxBufferSize = int.MaxValue;
            binding.MaxReceivedMessageSize = int.MaxValue;
            binding.TextEncoding = Encoding.UTF8;
            binding.TransferMode = TransferMode.Buffered;
            binding.MessageEncoding = WSMessageEncoding.Text;

            //CTP
            //TestCTP(args, binding);

            //Import
            //TestImport(args, binding);

            //CopyScenario
            //TestCopyScenario(args, binding);

            //GetScenarios
            //TestGetScenarios(args, binding);

            //DeleteScenario
            //TestDeleteScenario(args, binding);
        }

        #region CTP
        /// <summary>
        /// Test ExtraServices CTP call
        /// </summary>
        /// <param name="args"></param>
        /// <param name="a_binding"></param>
        private static void TestCTP(string[] args, BasicHttpBinding a_binding)
        {
            if (args == null || args.Length != 5)
            {
                Console.WriteLine("Need 5 parameters separated by spaces: URL username password ItemExternalId WarehouseExternalId");
                Environment.Exit(1);
            }

            string url = args[0];
            string username = args[1];
            string pwd = args[2];
            string item = args[3];
            string warehouse = args[4];
            if (string.IsNullOrEmpty(item))
            {
                item = "Item 1";
            }

            if (string.IsNullOrEmpty(warehouse))
            {
                warehouse = "Warehouse 1";
            }

            APSWebServiceRef.ExtraServicesWebServiceClient client = new APSWebServiceRef.ExtraServicesWebServiceClient(a_binding, new EndpointAddress(url));

            //Create a typed request
            APSWebServiceRef.CtpRequest ctpRequest = new APSWebServiceRef.CtpRequest();

            //Example CTP request
            List<APSWebServiceRef.CtpRequestLine> ctpRequestLines = new List<APSWebServiceRef.CtpRequestLine>();
            ctpRequestLines.Add(new APSWebServiceRef.CtpRequestLine()
            {
                LineId = 1,
                ItemExternalId = item,
                WarehouseExternalId = warehouse,
                RequiredQty = 100,
                NeedDate = DateTime.Now.Add(TimeSpan.FromDays(7))
            });

            //Connection settings
            ctpRequest.TimeoutMinutes = 3;
            ctpRequest.CtpRequestLines = ctpRequestLines.ToArray();
            ctpRequest.ReserveMaterialsAndCapacity = true;
            ctpRequest.SchedulingType = APSWebServiceRef.CtpDefsESchedulingType.Optimize;
            Write(ctpRequest);
            DateTime start = DateTime.Now;
            Console.WriteLine($"Request sent at '{start}'");

            //Sending a request requires the username and password of an APS user with appropriate permissions, generally an admin.
            var resp = client.CTP(username, pwd, ctpRequest);
            Console.WriteLine($"Response received after '{DateTime.Now.Subtract(start)}'");
            Write(resp);
            Console.WriteLine("Press any key to exit.");
            Console.ReadLine();
        }

        private static void Write(APSWebServiceRef.CtpResponse a_resp)
        {
            Console.WriteLine($"    Code: '{a_resp.ReturnCode}'");
            Console.WriteLine($"    ErrorMessage: '{a_resp.ErrorMessage}'");
            if (a_resp.CtpResponseLines == null)
            {
                Console.WriteLine($"    CtpResponseLines: null");
            }
            else
            {
                Console.WriteLine($"    CtpResponseLines.Length: {a_resp.CtpResponseLines.Length}");
                foreach (var respLine in a_resp.CtpResponseLines)
                {
                    Console.WriteLine($"        LineId: '{respLine.LineId}'");
                    Console.WriteLine($"        ScheduledStartDate: '{respLine.ScheduledStartDate}'");
                    Console.WriteLine($"        ScheduledStartDate: '{respLine.ScheduledEndDate}'");
                }
            }
        }

        private static void Write(APSWebServiceRef.CtpRequest a_req)
        {
            Console.WriteLine($"    Priority: '{a_req.Priority}'");
            Console.WriteLine($"    ReserveMaterialsAndCapacity: '{a_req.ReserveMaterialsAndCapacity}'");
            Console.WriteLine($"    SchedulingType: '{a_req.SchedulingType}'");
            Console.WriteLine($"    TimeoutMinutes: '{a_req.TimeoutMinutes}'");
            Console.WriteLine($"    CtpRequestLines.Length: '{a_req.CtpRequestLines.Length}'");
            foreach (var reqLine in a_req.CtpRequestLines)
            {
                Console.WriteLine($"        LineId: '{reqLine.LineId}'");
                Console.WriteLine($"        WarehouseExternalId: '{reqLine.WarehouseExternalId}'");
                Console.WriteLine($"        ItemExternalId: '{reqLine.ItemExternalId}'");
                Console.WriteLine($"        NeedDate: '{reqLine.NeedDate}'");
                Console.WriteLine($"        RequiredQty: '{reqLine.RequiredQty}'");
            }
        }
        #endregion CTP

        #region Import
        /// <summary>
        /// Test ExtraServices Import call
        /// </summary>
        /// <param name="a_args"></param>
        /// <param name="a_binding"></param>
        private static void TestImport(string[] a_args, BasicHttpBinding a_binding)
        {
            DateTime start = new DateTime();

            //Valiate args
            if (a_args == null || a_args.Length < 4)
            {
                Console.WriteLine("Need 4 required parameters separated by spaces: ExtraServicesURL username password IntefaceServiceURL, 3 optional parameters: ScenarioName ScenarioId CreateScenarioIfNew");
                Environment.Exit(1);
            }

            string extraServicesUrl = a_args[0];
            string username = a_args[1];
            string pwd = a_args[2];
            string interfaceUrl = a_args[3];
            string scenarioName = a_args[4];
            long scenarioId;
            if (!Int64.TryParse(a_args[5], out scenarioId) || scenarioId <= 0)
            {
                scenarioId = long.MinValue;
            }

            bool createScenarioIfNew;
            if (!Boolean.TryParse(a_args[6], out createScenarioIfNew))
            {
                createScenarioIfNew = false;
            }

            //Create client
            APSWebServiceRef.ExtraServicesWebServiceClient client = new APSWebServiceRef.ExtraServicesWebServiceClient(a_binding, new EndpointAddress(extraServicesUrl));

            //Validate Scenario info provided
            List<ScenarioConfirmation> matchingScenarios = new List<ScenarioConfirmation>();
            if (scenarioId != long.MinValue || !string.IsNullOrEmpty(scenarioName))
            {
                //Poll all scenarios for match
                GetScenariosResponse getScenariosResp = CreateAndSendGetScenariosRequest("", 5, username, pwd, client);
                foreach (ScenarioConfirmation scenario in getScenariosResp.Confirmations)
                {
                    if (scenario.ScenarioId == scenarioId || scenario.ScenarioName == scenarioName)
                    {
                        matchingScenarios.Add(scenario);
                    }
                }

                //Validate scenario match results
                if (matchingScenarios.Count == 2)
                {
                    //Scenario Id precedence
                    if (scenarioId != long.MinValue && matchingScenarios.Any(s => s.ScenarioId == scenarioId))
                    {
                        ScenarioConfirmation matchingScenario = matchingScenarios.Find(s => s.ScenarioId == scenarioId);
                        scenarioName = matchingScenario.ScenarioName;
                    }
                    else if (!string.IsNullOrEmpty(scenarioName) && matchingScenarios.Any(s => s.ScenarioName == scenarioName))
                    {
                        ScenarioConfirmation matchingScenario = matchingScenarios.Find(s => s.ScenarioName == scenarioName);
                        scenarioId = matchingScenario.ScenarioId;
                    }
                }
                else if (matchingScenarios.Count == 1)
                {
                    ScenarioConfirmation matchingScenario = new ScenarioConfirmation();

                    if (scenarioId != long.MinValue)
                    {
                        matchingScenario = matchingScenarios.Find(s => s.ScenarioId == scenarioId);
                        scenarioName = matchingScenario.ScenarioName;
                    }
                    else if (!string.IsNullOrEmpty(scenarioName))
                    {
                        matchingScenario = matchingScenarios.Find(s => s.ScenarioName == scenarioName);
                        scenarioId = matchingScenario.ScenarioId;
                    }
                }
                else if (createScenarioIfNew) //validation failed - Create New
                {
                    CopyScenarioResponse copyScenarioResp = CreateAndSendCopyScenarioRequest(scenarioId, scenarioName, createScenarioIfNew, username, pwd, client, ref start);
                    DisplayCopyScenarioResponse(copyScenarioResp);
                    if (copyScenarioResp.ReturnCode == EApsWebServicesResponseCodes.Success)
                    {
                        scenarioId = copyScenarioResp.Confirmation.ScenarioId;
                        scenarioName = copyScenarioResp.Confirmation.ScenarioName;
                    }
                }
                else //validation failed - Default (all scenarios)
                {
                    Console.WriteLine($"Note: Invalid scenario data entered -  import will perform on all scenarios");
                    scenarioId = long.MinValue;
                    scenarioName = "";
                }
            }

            //Create a Import request
            APSWebServiceRef.ImportRequest importRequest = new ImportRequest() { 
                InterfaceServiceUrl = interfaceUrl,
                ScenarioId = scenarioId,
                ScenarioName = scenarioName,
                CreateScenarioIfNew = createScenarioIfNew
            };
            importRequest.UserName = username;
            importRequest.Password = pwd;

            //Sending a request
            start = DateTime.Now;
            //ApsWebServiceResponseBase resp = client.Import(importRequest);
            //Console.WriteLine($"Response received after '{DateTime.Now.Subtract(start)}'");
            Task<ApsWebServiceResponseBase> importTask = client.ImportAsync(importRequest);
            ApsWebServiceResponseBase importResponse = importTask.Result;
            Trace($"Response received after '{DateTime.Now.Subtract(start)}'");
            Write(importResponse);
            Console.WriteLine("Press any key to exit.");
            Console.ReadLine();
        }

        private static void Trace(string a_message)
        {
            Console.WriteLine(a_message);

            File.AppendAllText(Path.Combine(Environment.CurrentDirectory, "log.txt"), a_message + Environment.NewLine);
        }

        private static void Write(ApsWebServiceResponseBase a_response)
        {
            Trace($"Operation code: {a_response.ResponseCode}");
            if (!a_response.Exception)
            {
                Trace($"Completed with Zero Errors");
            }
            else
            {
                Trace($"Completed with Errors");
                Trace(a_response.FullExceptionText);
            }
        }
        #endregion Import

        #region CopyScenario
        /// <summary>
        /// Test ExtraServices Import call
        /// </summary>
        /// <param name="a_args"></param>
        /// <param name="a_binding"></param>
        private static void TestCopyScenario(string[] a_args, BasicHttpBinding a_binding)
        {
            DateTime start = new DateTime();

            //Valiate args
            if (a_args == null || a_args.Length < 4)
            {
                Console.WriteLine("Need 3 required parameters separated by spaces: ExtraServicesURL username password, 3 optional parameters: ScenarioId ScenarioName CreateScenarioIfNew");
                Environment.Exit(1);
            }

            string extraServicesUrl = a_args[0];
            string username = a_args[1];
            string pwd = a_args[2];
            long scenarioId;
            if (!Int64.TryParse(a_args[3], out scenarioId) || scenarioId <= 0)
            {
                scenarioId = long.MinValue;
            }
            string scenarioName = a_args[4];

            bool createScenarioIfNew;
            if (!Boolean.TryParse(a_args[5], out createScenarioIfNew))
            {
                createScenarioIfNew = false;
            }

            APSWebServiceRef.ExtraServicesWebServiceClient client = new APSWebServiceRef.ExtraServicesWebServiceClient(a_binding, new EndpointAddress(extraServicesUrl));
            
            //Create and Send CopyScenarioRequest
            CopyScenarioResponse resp = CreateAndSendCopyScenarioRequest(scenarioId, scenarioName, createScenarioIfNew, username, pwd, client, ref start);

            //Display response stats
            Console.WriteLine($"Response received after '{DateTime.Now.Subtract(start)}'");
            Write(resp);

            //Display response data
            DisplayCopyScenarioResponse(resp);

            Console.WriteLine("Press any key to exit.");
            Console.ReadLine();
        }

        private static void DisplayCopyScenarioResponse(CopyScenarioResponse resp)
        {
            if (resp.ReturnCode == EApsWebServicesResponseCodes.Success)
            {
                Console.WriteLine($"Success: Scenario '{resp.Confirmation.ScenarioName}' created under Id:{resp.Confirmation.ScenarioId} with Type:{resp.Confirmation.ScenarioType}");
                if (resp.Exception)
                {
                    Console.WriteLine($"Errors: Invalid scenario data entered - {resp.ErrorMessage}");
                }
            }
            else
            {
                Console.Write("Scenario Copy not completed.");
            }
        }

        private static CopyScenarioResponse CreateAndSendCopyScenarioRequest(long a_scenarioId, string a_scenarioName, bool a_createScenarioIfNew, string a_username, string a_pwd, ExtraServicesWebServiceClient a_client, ref DateTime a_start)
        {
            //Create a typed request
            CopyScenarioRequest copyRequest = new CopyScenarioRequest()
            {
                ScenarioId = a_scenarioId,
                ScenarioName = a_scenarioName,
                CreateScenarioIfNew = a_createScenarioIfNew,
                TimeoutMinutes = 3
            };

            copyRequest.UserName = a_username;
            copyRequest.Password = a_pwd;

            //Sending a request
            a_start = DateTime.Now;
            CopyScenarioResponse resp = a_client.CopyScenario(copyRequest);
            return resp;
        }
        #endregion CopyScenario

        #region GetScenarios
        /// <summary>
        /// Test ExtraServices GetScenarios call for scenarios of specified type (or default all types)
        /// </summary>
        /// <param name="a_args"></param>
        /// <param name="a_binding"></param>
        private static void TestGetScenarios(string[] a_args, BasicHttpBinding a_binding)
        {
            DateTime start = new DateTime();

            //Valiate args
            if (a_args == null || a_args.Length < 3)
            {
                Console.WriteLine("Need 3 required parameters separated by spaces: ExtraServicesURL username password, 2 optional parameter: ScenarioType (ie. Whatif, Live, Published, RuleSeek, InsertJobs, Game, ShortTerm, Pruned), TimeoutMinutes");
                Environment.Exit(1);
            }

            string extraServicesUrl = a_args[0];
            string username = a_args[1];
            string pwd = a_args[2];
            string scenarioType = a_args[3];
            if (!double.TryParse(a_args[4], out double timeout))
            {
                timeout = 5;
            }

            APSWebServiceRef.ExtraServicesWebServiceClient client = new APSWebServiceRef.ExtraServicesWebServiceClient(a_binding, new EndpointAddress(extraServicesUrl));

            //Create and send GetScenariosRequest
            start = DateTime.Now;
            GetScenariosResponse resp = CreateAndSendGetScenariosRequest(scenarioType, timeout, username, pwd, client);

            //Display response stats
            Console.WriteLine($"Response received after '{DateTime.Now.Subtract(start)}'");
            Trace($"Response received after '{DateTime.Now.Subtract(start)}'");
            Write(resp);

            //Display response data
            if (resp.Confirmations.Length > 0)
            {
                Console.WriteLine($"{Environment.NewLine}{resp.Confirmations.Length} Current scenarios of Type '{(string.IsNullOrEmpty(scenarioType) ? "All Types" : scenarioType)}'{Environment.NewLine}");
                foreach (ScenarioConfirmation scenarioInfo in resp.Confirmations)
                {
                    Console.WriteLine($"  Name:{scenarioInfo.ScenarioName}{Environment.NewLine}    Id:{scenarioInfo.ScenarioId}{Environment.NewLine}  Type:{scenarioInfo.ScenarioType}{Environment.NewLine}");
                }
            }

            Console.WriteLine("Press any key to exit.");
            Console.ReadLine();
        }

        private static GetScenariosResponse CreateAndSendGetScenariosRequest(string a_scenarioType, double a_timeout, string a_username, string a_pwd, ExtraServicesWebServiceClient a_client)
        {
            //Create a typed request
            APSWebServiceRef.GetScenariosRequest getScenariosRequest = new GetScenariosRequest()
            {
                ScenarioType = a_scenarioType,
                TimeoutMinutes = a_timeout
            };
            getScenariosRequest.UserName = a_username;
            getScenariosRequest.Password = a_pwd;

            //Sending request
            return a_client.GetScenarios(getScenariosRequest);
        }
        #endregion

        #region DeleteScenario
        /// <summary>
        /// Test ExtraServices Delete Scenario call
        /// </summary>
        /// <param name="a_args"></param>
        /// <param name="a_binding"></param>
        private static void TestDeleteScenario(string[] a_args, BasicHttpBinding a_binding)
        {
            DateTime start = new DateTime();

            //Valiate args
            if (a_args == null || a_args.Length < 4)
            {
                Console.WriteLine("Need 4 total required parameters separated by spaces: ExtraServicesURL username password ScenarioId ScenarioName (Note:at least 1 Scenario identifier required, both can be submitted)");
                Environment.Exit(1);
            }

            string extraServicesUrl = a_args[0];
            string username = a_args[1];
            string pwd = a_args[2];
            long scenarioId;
            if (!Int64.TryParse(a_args[3], out scenarioId) || scenarioId <= 0)
            {
                scenarioId = long.MinValue;
            }
            string scenarioName = a_args[4];

            //Create client
            APSWebServiceRef.ExtraServicesWebServiceClient client = new APSWebServiceRef.ExtraServicesWebServiceClient(a_binding, new EndpointAddress(extraServicesUrl));
                        
            //Create and Send ScenarioDeleteRequest
            DeleteScenarioResponse resp = CreateAndSendScenarioDeleteRequest(scenarioId, scenarioName, username, pwd, client, ref start);

            //Display response stats
            Console.WriteLine($"Response received after '{DateTime.Now.Subtract(start)}'");
            Write(resp);

            //Display response data
            if (resp.ReturnCode == EApsWebServicesResponseCodes.Success)
            {
                Console.WriteLine($"Success: Scenario validated and deleted for '{resp.Confirmation.ScenarioName}' with Id '{resp.Confirmation.ScenarioId}'");

                if (resp.Exception)
                {
                    Console.WriteLine($"Input Errors: {resp.ErrorMessage}");
                }
            }
            else
            {
                Console.Write("Scenario Delete not completed - ");

                if (resp.ReturnCode == EApsWebServicesResponseCodes.InvalidLiveScenario)
                {
                    Console.WriteLine($"Attempting to delete Live Scenario - Id '{scenarioId}' received");
                }
                else if (resp.ReturnCode == EApsWebServicesResponseCodes.InvalidScenarioIdAndName)
                {
                    Console.WriteLine($"No valid scenario data entered");
                }
                else
                {
                    Console.WriteLine($"{resp.ErrorMessage}");
                }

                if (resp.Exception)
                {
                    Console.WriteLine($"Input Errors: {resp.ErrorMessage}");
                }
            }

            Console.WriteLine("Press any key to exit.");
            Console.ReadLine();
        }

        private static DeleteScenarioResponse CreateAndSendScenarioDeleteRequest(long a_scenarioId, string a_scenarioName, string a_username, string a_pwd, ExtraServicesWebServiceClient a_client, ref DateTime a_start)
        {
            //Create a typed request
            DeleteScenarioRequest deleteScenarioRequest = new DeleteScenarioRequest()
            {
                ScenarioId = a_scenarioId,
                ScenarioName = a_scenarioName,
                TimeoutMinutes = 3
            };

            deleteScenarioRequest.UserName = a_username;
            deleteScenarioRequest.Password = a_pwd;

            //Sending a request
            a_start = DateTime.Now;
            DeleteScenarioResponse resp = a_client.DeleteScenario(deleteScenarioRequest);
            return resp;
        }
        #endregion DeleteScenario
    }
}
